################################################################################### #
#
# This function is inspired in a script for the function "fitDistr" from the R package
# "propagate" created by Andrej-Nikolai Spiess <a.spiess@uke.uni-hamburg.de>
#
# Andrej-Nikolai Spiess (2014). propagate: Propagation of Uncertainty.
# R package version 1.0-4. http://CRAN.R-project.org/package=propagate
#
# Some parts or ideas of the script for function "fitDistr" are used, but  
# here we to estimate CDF and not the PDF in the case of "fitDistr". A more informative
# plot is also incorporated. The studentized residuals are provided as well.
#
# Robersy Sanchez Dec. 15, 2015
################################################################################### #

fitCDF <- function (object, verbose = TRUE, plot = TRUE, plot.num = 1, dist.names,...) 
{
  # plot.num is the number of istributions to be plotted
  # dist.names: if one or distributions are selected,
  # then only those distribution are analized. All the follwiing
  # distributions are analyzed:
  #   "Normal", "Log-normal", "Generalized Normal","Laplace", "Gamma",
  #   "3P Gamma", "Generalized Gamma", "Weibull", "3P Weibull", "Beta", 
  #   "3P Beta", "4P Beta", "Generalized Beta", "Rayleigh", "Exponential",
  #   "2P Exponential"
  # After return the plots, a list with following values is provided:
  #   aic: Akaike information creterion
  #   fit: list of results of fitted distribution, with parameter values
  #   bestfit: the best fitted distribution according to AIC 
  #   fitted: fitted values from the best fit
  #   rstudent: studentized residuals
  #   residuals: residuals
  
  require( minpack.lm )
  require( numDeriv )
  
  # "count" function from "propagate"
  counter <- function (i) 
  {
    if (i%%10 == 0) 
      cat(i)
    else cat(".")
    if (i%%50 == 0) 
      cat("\n")
    flush.console()
  }
  
  nls.rstudent <- function( grad, residuals, num.par )
  {
    # *** Studentized residuals for nls *** #
    # grad: derivative of model as obtained from function deriv
    # R.S
    n = length( residuals )
    h = diag( grad %*% solve( crossprod( grad ) ) %*% t( grad ) )
    s = sqrt( n*var( residuals, na.rm = T )/( n - num.par ) )
    residuals/( s * sqrt( 1 - h ) ) 
  }
  
  options(warn = -1)
  if (is.vector(object)) 
    X <- sort( object )
  else stop( "object must be a numeric vector!" )
  MEAN <- mean( X, na.rm = TRUE)
  VAR <- var( X, na.rm = TRUE)
  SD <- sd(X, na.rm = TRUE)
  MIN <- min( X, na.rm = TRUE)
  MAX <- max( X, na.rm = TRUE)
  Q = unname( quantile( X, 0.632, na.rm = TRUE ) )
  Fy = ecdf( X )
  pX = Fy( X ) 
    
  # *** Generalized normal CDF *** #
  # https://en.wikipedia.org/wiki/Generalized_normal_distribution
  pgnorm <- function( q , mean = 0, alpha = 1, beta = 1 )
  {
   y = ( abs( q - mean )/alpha )^beta
   1/2 + sign( q - mean ) * pgamma( y, 1/beta )/( 2*gamma( 1/beta ) )
  }
  
    
  # *** Definition of Laplace CDF *** #
  plaplace <- function( q , mean = 0, sigma = 1 )
  {
   1/sqrt( 2 ) + 1/sqrt( 2 ) * ( 1 - exp( -abs( q - mean )/sigma ) )
  } 
  
  # *** Definition of 3P gamma CDF *** #
  pgamma3 <- function( q, shape, rate, mu ) pgamma( q - mu, shape, rate )
  
  # *** Definition of generalized gamma distribution *** #
  pggamma <- function ( q, alpha = 1, scale = 1, mu = 0, psi = 1, lower.tail = TRUE, log.p = FALSE ) 
  {
    #   Hand-book on  STATISTICAL DISTRIBUTIONS for experimentalists ( pag 73 )
    #   by Christian Walck
    #   Particle Physics Group Fysikum
    #   University of Stockholm (e-mail: walck@physto.se )
    y <- ( ( q - mu )/scale )^alpha
    # See section Note at ?pgamma
    p <- pgamma( y, psi, lower.tail = lower.tail, log.p = log.p )
    p[ alpha < 0 ] <- NaN
    p
  }
  
  # *** Generalized beta CDF *** #
  pgbeta <- function ( q, shape1, shape2, lambda = 1, lower.tail = TRUE, log.p = FALSE ) 
  {
    ans <- pbeta( q = 1/(1 + (1/q - 1)/lambda ), shape1 = shape1, 
                 shape2 = shape2, lower.tail = lower.tail, log.p = log.p )
    ans[ lambda <= 0 ] <- NaN
    ans
  }
  
  # *** 3P beta CDF *** #
  pbeta3 <- function (q, shape1 = 2, shape2 = 3, a = 0, lower.tail = TRUE, log.p = FALSE )
  {
    pbeta( q - a, shape1, shape2, lower.tail = lower.tail, log.p = log.p ) 
  }
  
  # *** 4P beta CDF *** #
  pbeta4 <- function (q, shape1 = 2, shape2 = 3, a = 0, b = 1, lower.tail = TRUE, log.p = FALSE )
  {
    pbeta( ( q - a )/( b - a ), shape1, shape2, lower.tail = lower.tail, log.p = log.p ) 
  }
  
  # *** Definition of 3P Weibull *** #
  pweibull3 <- function( q, shape, scale, mu ) pweibull( q - mu, shape, scale )
  
  # *** Definition of Rayleigh distribution *** #
  prayleigh <- function( q, sigma )  1 - exp( -q^2/( 2 * sigma^2 ) )
  
  # *** Definition of 2P Exponential *** #
  pexp2 <- function( q, rate, mu ) pexp( q - mu , rate = 1, lower.tail = TRUE, log.p = FALSE )
  
  
  distNAMES <- c( "Normal", "Log-normal", "Generalized Normal","Laplace", "Gamma",
                  "3P Gamma", "Generalized Gamma", "Weibull", 
                  "3P Weibull", "Beta", "3P Beta", "4P Beta",
                  "Generalized Beta", "Rayleigh", "Exponential", "2P Exponential" )
  
  funLIST <- list( pnorm, plnorm, pgnorm, plaplace, pgamma, pgamma3, pggamma, pweibull, pweibull3, pbeta,
                   pbeta3, pbeta4, pgbeta, prayleigh, pexp, pexp2 )
  
  parLIST <- list( norm = c( mean = MEAN, sd = SD ), 
                   lnorm = c( meanlog = mean( log1p( X ), na.rm = TRUE ) ), 
                   gnorm = c( mean = MEAN, alpha = 1, beta = 1 ),
                   laplace = c( mean = MEAN, sigma = sqrt( VAR ) ), 
                   gamma = c( shape = MEAN^2/VAR, rate = MEAN/VAR ), 
                   gamma3 = c( shape = MEAN^2/VAR, rate = MEAN/VAR, mu = 0 ),
                   ggamma = c( alpha = MEAN^2/VAR, scale = VAR/MEAN, mu = MIN, psi = 1 ), 
                   weibull = c( shape = log( 2 ), scale = Q ),
                   weibull3 = c( mu = MIN, shape = log( 2 ), scale = Q ),
                   beta = c( shape1 = 1, shape2 = 2 ),
                   beta3 = c( shape1 = 1, shape2 = 2, a = MIN ),
                   beta4 = c( shape1 = 2, shape2 = 3, a = 0.9 * MIN, b = 1.1 * MAX ), 
                   gbeta = c( shape1 = 1, shape2 = 2, lambda = 1 ), 
                   rayleigh = c( sigma = SD ),
                   exp = c( rate = 1 ),
                   exp2 = c( rate = 1, mu = 0 )
                   )
 
  if( missing( dist.names ) ) dist.names = distNAMES
  if( dist.names != length( distNAMES ) )
  {
    ind = as.numeric( na.omit( match( dist.names, distNAMES ) ) )
    distNAMES = distNAMES[ ind ]
    funLIST = funLIST[ ind ]
    parLIST = parLIST[ ind ]
  }
  
  optFun <- function( par, probfun, quantiles, prob, eval = FALSE ) {
    START <- as.list( par )
    START$q <- quantiles
    EVAL <- try(do.call( probfun, START ), silent = TRUE)
    if (inherits( EVAL, "try-error" ) ) 
      return( NA )
    EVAL[ is.nan( EVAL ) ] <- 0
    RSS <- ( prob - EVAL )^2
    if( eval ) 
      return( EVAL )
    else return( RSS )
  }
  
  fitAIC <- function( fitobj ) {
    RESID <- fitobj$fvec
    sse = sum( RESID^2, na.rm = TRUE )
    N <- length( RESID )
    N * ( 1 + log( 2 * pi ) + log( sse/N ) ) + 2 * ( 1L + length( fitobj$par ) )
  }
  
  fitLIST <- vector("list", length = length(distNAMES))
  AICS <- rep(NA, length(distNAMES))
 
   for (i in 1:length(distNAMES)) {
    if (verbose) 
      cat("Fitting ", distNAMES[i], " distribution..", 
          sep = "")
    GRID <- matrix(parLIST[[i]], nrow = 1)
    colnames(GRID) <- names(parLIST[[i]])
    rssVEC <- rep(NA, nrow(GRID))
    for (j in 1:nrow(GRID)) {
      if (verbose) 
        counter(j)
      PARS <- GRID[j, ]
      FIT <- try( nls.lm( par = PARS, fn = optFun, probfun = funLIST[[i]], 
                          quantiles = X, prob = pX,
                          control = nls.lm.control( maxiter = 10^4, maxfev = 1e+5, ptol = 1e-12 ) ),
                  silent = TRUE )
      if (inherits(FIT, "try-error")) 
        rssVEC[j] <- NA
      else rssVEC[j] <- FIT$deviance
    }
    if ( rssVEC[ j ] > 1 ) {
      WHICH <- which.min( rssVEC[ j ] )
      bestPAR <- GRID[WHICH, ]
      FIT <- try( nls.lm( par = bestPAR, fn = optFun, probfun = funLIST[[i]], 
                          quantiles = X, prob = pX, 
                          control = nls.lm.control( maxiter = 10^4, maxfev = 1e+5, ptol = 1e-12 ) ),
                  silent = TRUE)
    }
    if (inherits(FIT, "try-error")) {
      FIT <- NA
      if (verbose) 
        cat("Error!\n")
    }
    else {
      fitLIST[[i]] <- FIT
      AICS[i] <- tryCatch( fitAIC( FIT ), error = function(e) NA )
      if (verbose) 
        cat("Done.\n")
    }
  }
  ORDER <- order(AICS)
  aicDAT <- data.frame( Distribution = distNAMES, AIC = AICS)
  aicDAT <- aicDAT[ ORDER, ]
  
  j = 1
  for( k in min( plot.num, length( dist.names ) ):1 )
  {
    cat( " * Estimating Studentized residuals for distribution # ", j, "\n" )
    SEL <- ORDER[ k ]
    bestFIT <- fitLIST[[SEL]]
    evalLIST <- as.list(bestFIT$par)
    evalLIST$q <- X
    evalY <- do.call(funLIST[[SEL]], evalLIST)
    
    #   # Derivative of the best fit function
    evalLST <- as.list( bestFIT$par )
    evalLST$func <- funLIST[[SEL]]
    evalLST$x <- X
    evalLST$method <- "simple"
    gradient <- do.call( numDeriv:::grad, evalLST )
    
    # NaN/missing values are replace by cubic spline interpolation
    grad.spline = splinefun( X, gradient )
    ind = which( is.na( gradient ) )
    gradient[ ind ] <- grad.spline( X[ ind ] )
    residuals = pX - evalY
    
    rstudent = nls.rstudent( gradient, residuals, length( bestFIT$par ) )
    
    if( plot )
    {
      cat( " * Plot # ", j, "...\n" )
      par( mfrow = c( 2, 2 ), mar = c( 2, 2.5, 2.2, 1 ) + 0.2, mgp = c( 1.2, 0.4, 0 ) )
      plot( Fy, verticals = TRUE, col = "blue", pch = 20, xlab = expression( italic( "x" ) ),
            ylab = "CDF", main =  paste( distNAMES[ SEL ], "Distribution" ), cex.main = 0.9 )
      grid( NULL,NULL, lty = 6, col = "cornsilk2" )
      lines( evalLIST$q, evalY, col = 2, lty = 2, lwd = 2 )
      mtext( text = paste( "AIC =", round( AICS[ SEL ], 3 ) ), cex = 0.6  )
      
      ## PP-plot
      par( mar = c( 2, 2, 2.2, 1 ) + 0.2, mgp = c( 1.2, 0.4, 0 ) )
      plot( pX, evalY, col = "blue" , main = "P-P Plot", pch = 20, cex = 0.4, xlab = "Empirical CDF",
            ylab = "Theoretical CDF", cex.main = 0.9 )
      grid( NULL,NULL, lty = 6, col = "cornsilk2" )
      abline( 0, 1, col= "red", lwd = 2 ) # Reference line y = x
      pars = bestFIT$par
      pars = t( cbind( names( pars ), format( round ( pars, 3 ), 3 ) ) )
      mtext( text = paste( pars, collapse = " " ), cex = 0.6 )
      
      par( mar = c( 3.5, 3, 1, 1 ) + 0.07, mgp = c( 1.2, 0.4, 0 ) )
      plot( X, rstudent, pch = 20, xlab = expression( italic( "x" ) ), 
            ylab =  "Studentized residuals", col = "blue", cex = 0.4, cex.main = 0.9 )
      grid( NULL,NULL, lty = 6, col = "cornsilk2" )
      abline( h = 2, col= "red", lwd = 2, lty=2 ) # Reference line y = x
      abline( h = -2, col= "red", lwd = 2, lty=2 ) # Reference line y = x
      
      par( mar = c( 3.5, 2, 1, 1 ) + 0.07, mgp = c( 1.2, 0.4, 0 ) )
      qqnorm( rstudent,  col = "blue" , pch = 20, cex = 0.4, cex.main = 0.9 )
      grid( NULL,NULL, lty = 6, col = "cornsilk2" )
      qqline( rstudent, col= "red", lwd = 2 ) # Reference line y = x
      j = j + 1
    }
  }
  names(fitLIST) <- distNAMES
  fitLIST = fitLIST[ as.character( aicDAT$Distribution ) ]
  cat( "** Done ***\n" )
  return( list( aic = aicDAT, fit = fitLIST, bestfit = bestFIT, 
                fitted = evalY, rstudent = rstudent, residuals = residuals ) )
}
