####------------------------------------- ------------------------- ###
#0
# This a homemade function to illustrate the failure of Q-Qnorm plot
# Kolmogorov-Smirnov test with problems for large data sets and ways to 
# solve it.
# Robersy Sanchez
# Last modified 01/232016
#### --------------------------------------------------------------- ###

qq.weibull<-function( s = 1, m = 3, perturb = NULL, factor = 1, alpha = 0.6831651,
                      scale = 2.5114992, mu = 0.01, plot = TRUE, num.permt = NULL, seed = 1223 )
{
  set.seed( seed ) # randomm seed
  require( FAdist )
  require( numDeriv )

  rstudentized <- function( grad, residuals, num.par )
  {
    # *** Studentized residuals for nls *** #
    # grad: derivative of model as obtained from function deriv
    # R.S
    n = length( residuals )
    h = diag( grad %*% solve( crossprod( grad ) ) %*% t( grad ) )
    s = sqrt( n*var( residuals, na.rm = T )/( n - num.par ) )
    residuals/( s * sqrt( 1 - h ) ) 
  }
  
  ## theorical distribution
  qw <- function(x) qweibull3( x , shape = alpha, scale= scale, thres = mu )
  pw <- function(x) pweibull3( x , shape = alpha, scale= scale, thres = mu )
  rw <- function( x, sh = alpha, sc = scale, th = mu ) rweibull3( x , shape = sh, scale= sc, thres = th )
  
  x.teo <- rw( s * 10^m ) ## theorical quantiles
  x.emp <- jitter( x.teo, factor = factor, amount = perturb ) ## Empirical quantiles, small perturbation
  
  p.teo = pw( x.emp ) # Probabilities predicted for the empirical data, according to the model
  Fy = ecdf( x.emp ) # Empirical Cumulative Distribution Function
  p.emp = Fy( x.emp )
  
  #   # Derivative of the best fit function
  evalLST <- list( shape = alpha, scale = scale, thres = mu )
  evalLST$func <- pweibull3
  evalLST$x <- x.emp
  evalLST$method <- "simple"
  gradient <- do.call( numDeriv:::grad, evalLST )
  
  # NaN/missing values are replace by cubic spline interpolation
  grad.spline = splinefun( x.emp, gradient )
  ind = which( is.na( gradient ) )
  gradient[ ind ] <- grad.spline( x.emp[ ind ] )
  residuals = p.emp - p.teo
  
  rstudent = rstudentized( gradient, residuals, 3 )
  
  if( plot )
  {
    par( mfrow = c( 2, 2 ), mar = c( 2, 2.5, 2.2, 1 ) + 0.2, mgp = c( 1.2, 0.4, 0 ) )
    qqplot( x.teo, x.emp, main="QQ-plot. 3P Weibull Dist",
            xlab = "Theoretical distribution", ylab = "Empirical distribution",
            col = "blue", pch = 20, cex.main = 0.9 ) ## QQ-plot
    qqline( x.emp, distribution = qw, lty = 2, lwd = 2 ) # Reference line
    grid( NULL,NULL, lty = 6, col = "cornsilk2" )
    
    ## PP-plot
    par( mar = c( 2, 2, 2.2, 1 ) + 0.2, mgp = c( 1.2, 0.4, 0 ) )
    plot( p.teo, p.emp, main="PP-plot. 3P Weibull Dist",
          xlab = "Theoretical CDF", ylab = "Empirical CDF",
          col = "blue", pch = 20, cex.main = 0.9)
    grid( NULL,NULL, lty = 6, col = "cornsilk2" )
    abline(0,1) # Reference line y = x
    
    par( mar = c( 3.5, 3, 1, 1 ) + 0.07, mgp = c( 1.2, 0.4, 0 ) )
    plot( x.emp, rstudent, pch = 20, xlab = expression( italic( "x" ) ), 
          ylab =  "Studentized residuals", col = "blue", cex = 0.4, cex.main = 0.9 )
    grid( NULL,NULL, lty = 6, col = "cornsilk2" )
    abline( h = 2, col= "red", lwd = 2, lty=2 ) # Reference line y = x
    abline( h = -2, col= "red", lwd = 2, lty=2 ) # Reference line y = x
    
    par( mar = c( 3.5, 2, 1, 1 ) + 0.07, mgp = c( 1.2, 0.4, 0 ) )
    qqnorm( rstudent,  col = "blue" , pch = 20, cex = 0.4, cex.main = 0.9 )
    grid( NULL,NULL, lty = 6, col = "cornsilk2" )
    qqline( rstudent, col= "red", lwd = 2 ) # Reference line y = x
  }
  
  if( missing( num.permt ) )
  {
    return( list( lm = summary( lm( p.teo ~ p.emp ) ), # gression summary
                  ks = ks.test( x.emp, pw ) ) ) # Kolmogorov-Smirnov Tests  
  }else{
    cat( "*** Performing permutation test for KS statistic. ", num.permt, " permutations...\n"  )
    # From Alastair Sanderson web page: Using R to analyse data
    # statistical and numerical data analysis with R
    # http://www.sr.bham.ac.uk/~ajrs/R/r-analyse_data.html
    PermTest <- function(x, y, R=999, testfun=ks.test ) {
      z <- c(x, y)  # pooled sample
      myfun <- function(a, b) suppressWarnings(unname(testfun(a, b)$statistic))
      DoIt <- function() {
        i <- sample(length(z), length(x))
        myfun(z[i], z[-i])  # z[-i] is everything *except* the "i" elements of z
      }
      pstats <- replicate( R, DoIt() )
      stat <- myfun(x, y)
      c( "p-value" = mean( c( stat, pstats ) >= stat ) )
    }
    
    p.perm = PermTest( x.emp, x.teo, R = num.permt )
    
    return( list( lm = summary( lm( p.teo ~ p.emp ) ), # regression summary
                  ks = ks.test( x.emp, pw ),
                  Permutation.p.value = p.perm ) )
  }
}
